#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define N_CHAR 400
#include "codes.h"
#include "get_para_pop_dyn.h"

static void help(char *);
static int
Read_parameters(char *FILE_IN, char *file_pdb, char chain[],
		char *file_seq, char *file_str, float *TEMP,
		float *sU1, float *sC0, float *sC1, int *REM, float *SEC_STR,
		int *REMUT, long *IT_MAX, int *NEUTRAL, int *MEANFIELD,
		int *OPT_LAMBDA, float *LAMBDA, float *DG_OPT, int *N_pop, 
		int *GET_FREQ, float *freq_nuc, float *trans_ratio,
		float *kCpG, float *TWONUC, char *EXCHANGE, char *MATRIX,
		int *FORMAT, int *PRINT_E, char **file_ali, char *MODEL,
		char *dir_out);
static int Find_string(char *flag, char **argv, int n_arg,int m,char *string);


int Get_para(int argc, char **argv,
	     char *file_pdb, char chain[],
	     char *file_seq, char *file_str, float *TEMP,
	     float *sU1, float *sC0, float *sC1, int *REM, float *SEC_STR,
	     int *REMUT, long *IT_MAX, int *NEUTRAL, int *MEANFIELD,
	     int *OPT_LAMBDA, float *LAMBDA, float *DG_OPT, int *N_pop, 
	     int *GET_FREQ, float *freq_nuc, float *trans_ratio,
	     float *kCpG, float *TWONUC, char *EXCHANGE, char *MATRIX,
	     int *FORMAT, int *PRINT_E, char **file_ali, char *MODEL,
	     char *dir_out)
{
  char FILE_IN[N_CHAR], string[N_CHAR];
  int r, i, j, i_nuc; float sum=0, p;

  if(Find_string("-h", argv, argc, 2, string))help(argv[0]);
  if(argc < 2){
    printf("ERROR, either input file or options must be specified\n");
    help(argv[0]);
  }

  strcpy(FILE_IN, argv[1]);
  file_pdb[0]='\0';
  r=Read_parameters(FILE_IN, file_pdb, chain, file_seq, file_str,
		    TEMP, sU1, sC0, sC1, REM, SEC_STR, REMUT, IT_MAX,
		    NEUTRAL, MEANFIELD, OPT_LAMBDA, LAMBDA, DG_OPT, N_pop, 
		    GET_FREQ, freq_nuc, trans_ratio, kCpG, TWONUC,
		    EXCHANGE, MATRIX, FORMAT, PRINT_E, file_ali, MODEL,
		    dir_out);

  /*********************************************/
  if(argc <= 2){
    if(r==0){
      printf("ERROR, input file %s does not exist\n", argv[1]);
      help(argv[0]);
    }else{
      goto check;
    }
  }
  printf("! Changing parameters with command line:\n");

  char w[80];
  for(j=2; j<argc; j++){
    if(strncmp(argv[j], "-pdb", 4)==0){
      j++; sscanf(argv[j], "%s", file_pdb);
    }else if(strncmp(argv[j], "-chain", 3)==0){
      j++; sscanf(argv[j], "%s", chain);
    }else if(strncmp(argv[j], "-pop", 4)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p >=1){
	*N_pop=p; printf("Population size: %d\n", *N_pop); j++;
      }
    }else if(strncmp(argv[j], "-seq", 4)==0){
      strcpy(file_seq, argv[j+1]); 
      printf("Sequence file: %s\n", file_seq);
    }else if(strncmp(argv[j], "-tt", 3)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p>1){
	*trans_ratio=p; j++;
	printf("Transition-transversion ratio: %.2f\n", *trans_ratio);
      }
    }else if(strncmp(argv[j], "-meanfield", 10)==0){
      *MEANFIELD=1;
      printf("Meanfield computation of substitution rates\n");
    }else if(strncmp(argv[j], "-model", 6)==0){
      j++; sscanf(argv[j], "%s", w);
      if((strcmp(w, "NAT")==0)||
	 (strcmp(w, "ALL")==0)||
	 (strcmp(w, "DG")==0)){
	strcpy(MODEL, w);
      }else{
	printf("WARNING, %s is not an allowed model. ", w);
	printf("Allowed models are NAT ALL DG\n");
	printf("Using default: %s\n", MODEL);
      }
    }else if(strncmp(argv[j], "-lambda", 7)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p>0){
	*LAMBDA=p; printf("Lambda parameter for meanfield: %3f\n", p); j++;
	*OPT_LAMBDA=0;
      }
    }else if(strncmp(argv[j], "-temp", 5)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p>0){
	*TEMP=p; printf("Temperature: %3f\n", p); j++;
      }
    }else if(strncmp(argv[j], "-sU1", 3)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p>0){
	*sU1=p; printf("Unfolded entropy per residue: %3f\n", *sU1); j++;
      }
    }else if(strncmp(argv[j], "-it", 3)==0){
      sscanf(argv[j+1], "%f", &p);
      if(p>10){
	*IT_MAX=p; printf("Substitutions per residue: %ld\n", *IT_MAX); j++;
      }
    }else if(strncmp(argv[j], "-ali", 4)==0){
      *file_ali=malloc(200*sizeof(char));
      sscanf(argv[j+1], "%s", *file_ali);
    }else if(strncmp(argv[j], "-freq", 5)==0){
      j++;
      if(strcmp(w, "nuc")==0){*GET_FREQ=1;}
      else if(strcmp(w, "aa")==0){*GET_FREQ=2;}
      else if(strcmp(w, "input")==0){*GET_FREQ=0;}
      else{
	printf("WARNING, %s is not an allowed fit for frequencies. ", w);
	printf("Allowed fits are aa, nuc, input\n");
	printf("Using default GET_FREQ= %d\n", *GET_FREQ);
      }
    }else if(strncmp(argv[j], "-f", 2)==0){
      sscanf(argv[j]+2, "%s", w); i_nuc=Code_nuc(w[0]);
      if((i_nuc<0)||(i_nuc>3)){
	printf("WARNING, %c is not an allowed nucleotideºn", w[0]);
	continue;
      }
      sscanf(argv[j+1], "%f", &p);
      if((p>0)&&(p<1)){
	freq_nuc[i_nuc]=p; j++;
	printf("freq_nuc(%c): %f\n", NUC_CODE[i_nuc], freq_nuc[i_nuc]);
      }
    }else if(strncmp(argv[j], "-gc", 3)==0){
      sscanf(argv[j+1], "%f", &p);
      if((p>0)&&(p<1)){
	freq_nuc[Code_nuc('G')]=0.5*p;
	freq_nuc[Code_nuc('C')]=0.5*p;
	freq_nuc[Code_nuc('T')]=0.5*(1-p);
	freq_nuc[Code_nuc('A')]=0.5*(1-p);
	for(i=0; i<4; i++)
	  printf("freq_nuc(%c): %3f\n", NUC_CODE[i], freq_nuc[i]);
	j++;
      }
      /*}else if(strncmp(argv[j], "-gencode", 8)==0){
      strcpy(file_code, argv[j+1]); j++;
      printf("Genetic code: %s\n", file_code);*/
    }else if(strncmp(argv[j], "-d", 2)==0){
      strcpy(dir_out,argv[j+1]); j++;
      printf("Directory in is: %s\n",dir_out);
    }else{
      printf("WARNING, unrecognized option %s\n", argv[j]);
    }
  }

  printf("! End changing parameters\n\n");

  /*********************************************/

  // Checks
 check:
  if(file_pdb[0]=='\0'){
    printf("ERROR, PDB file not specified\n");
    help(argv[0]);
  }
  if((*NEUTRAL==1)&&(*MEANFIELD==1)){
    printf("WARNING: MEANFIELD and NEUTRAL cannot be simultaneosuly true, ");
    printf("setting NEUTRAL=0\n");
    *NEUTRAL=0;
  }

  if((*N_pop < 1)&&(*NEUTRAL==0)&&(*MEANFIELD==0)){
    printf("ERROR, %d is too small population size\n", *N_pop);
    printf("Check the input file %s\n", FILE_IN);
  }

  for(i=0; i<4; i++)sum+=freq_nuc[i];
  if(sum!=1){
    printf("Warning, nucleotide frequencies are not normalized.\n");
    printf("Normalizing\n");
  }
  for(i=0; i<4; i++)freq_nuc[i]/=sum;
  return(0);
}

int Read_parameters(char *FILE_IN, char *file_pdb, char *chain,
		    char *file_seq, char *file_str,
		    float *TEMP, float *sU1, float *sC0, float *sC1,
		    int *REM, float *SEC_STR, int *REMUT,
		    long *IT_MAX, int *NEUTRAL, int *MEANFIELD,
		    int *OPT_LAMBDA, float *LAMBDA, float *DG_OPT, int *N_pop, 
		    int *GET_FREQ, float *freq_nuc, float *trans_ratio,
		    float*kCpG, float *TWONUC, char *EXCHANGE,
		    char *MATRIX, int *FORMAT, int *PRINT_E,
		    char **file_ali, char *MODEL, char *dir_out)
{
  FILE *file_in=fopen(FILE_IN, "r");
  char string[1000], nuc[3], dumm[80];
  int i_nuc; float p, sum=0;

  if(file_in==NULL){
    printf("WARNING, input file %s does not exist\n", FILE_IN);
    return(0);
  }
  printf("Reading parameters in %s\n", FILE_IN);

  char READ[100];
  while(fgets(string, sizeof(string), file_in)!=NULL){
    if(string[0]=='#')continue;
    // Protein files
    if(strncmp(string, "PDB=", 4)==0){
      sscanf(string+4,"%s", file_pdb);
    }else if(strncmp(string, "CHAIN=", 6)==0){
      sscanf(string+6,"%s", dumm); *chain=dumm[0];
    }else if(strncmp(string, "SEQ=", 4)==0){
      sscanf(string+4,"%s", READ);
      if(strncmp(READ, "NULL", 4)==0)continue;
      strcpy(file_seq, READ);
    }else if(strncmp(string, "ALI=", 4)==0){
      sscanf(string+4,"%s", READ);
      if(strncmp(READ, "NULL", 4)==0)continue;
      *file_ali=malloc(100*sizeof(char));
      strcpy(*file_ali, READ);
      // Thermodynamic parameters:
    }else if(strncmp(string, "TEMP=", 5)==0){
      sscanf(string+5,"%f", TEMP);
    }else if(strncmp(string, "SU1=", 3)==0){
      sscanf(string+4,"%f", sU1); 
    }else if(strncmp(string, "SC1=", 4)==0){
      sscanf(string+4,"%f", sC1); 
    }else if(strncmp(string, "SC0=", 4)==0){
      sscanf(string+4,"%f", sC0); 
    }else if(strncmp(string, "REM=", 4)==0){
      int dum;
      sscanf(string+4,"%d", &dum);
      if((dum<0)||(dum>3)){
	printf("WARNING, the variable REM can only be set to 1, 2 or 3\n");
      }else{
	*REM=dum;
      }
    }else if(strncmp(string, "A_LOC", 5)==0){
      sscanf(string+6,"%f", SEC_STR);
    }else if(strncmp(string, "REMUT", 5)==0){
      sscanf(string+6,"%d", REMUT);
    }else if(strncmp(string, "FILE_STR=", 9)==0){
      sscanf(string+9,"%s", file_str);
      // Selection model:
    }else if(strncmp(string, "TMAX=", 5)==0){
      sscanf(string+5,"%ld", IT_MAX);
    }else if(strncmp(string, "NEUTRAL=", 8)==0){
      sscanf(string+8,"%d", NEUTRAL);
    }else if(strncmp(string, "MEANFIELD=", 10)==0){
      sscanf(string+10,"%d", MEANFIELD);
    }else if(strncmp(string, "OPT_LAMBDA=", 11)==0){
      sscanf(string+11,"%d", OPT_LAMBDA);
    }else if(strncmp(string, "MODEL=", 6)==0){
      char w[80]; sscanf(string+6, "%s", w);
      if((strcmp(w, "NAT")!=0)&&
	 (strcmp(w, "ALL")!=0)&&
	 (strcmp(w, "DG")!=0)){
	printf("WARNING, %s is not an allowed model. ", w);
	printf("Allowed models are NAT ALL DG\n");
	printf("Using default: %s\n", MODEL);
      }else{
	strcpy(MODEL, w);
      }
    }else if(strncmp(string, "DG_OPT=", 7)==0){
      sscanf(string+7,"%f", DG_OPT);
    }else if(strncmp(string, "LAMBDA=", 7)==0){
      sscanf(string+7,"%f", LAMBDA);
      //printf("Lambda parameter for meanfield: %1f\n", *LAMBDA);
    }else if(strncmp(string, "NPOP=", 5)==0){
      sscanf(string+5,"%d", N_pop);
      // Mutation model
    }else if(strncmp(string, "GET_FREQ=", 9)==0){
      sscanf(string+9,"%d", GET_FREQ);
    }else if(strncmp(string, "FREQ", 4)==0){
      sscanf(string+4,"%s%f", nuc, &p);
      i_nuc=Code_nuc(nuc[0]); freq_nuc[i_nuc]=p; sum+=p;
    }else if(strncmp(string, "TT_RATIO=", 9)==0){
      sscanf(string+9,"%f", trans_ratio);
    }else if(strncmp(string, "PRINT_E", 7)==0){
      sscanf(string+8,"%d", PRINT_E);
    }else if(strncmp(string, "FORMAT=", 7)==0){
      char format[80];
      sscanf(string+7,"%s", format);
      if(strncmp(format, "PAML", 4)==0){
	*FORMAT=1;
      }else{
	printf("WARNING, non valid option FORMAT= %s\n", format);
      }
    }else if(strncmp(string, "EXCHANGE=", 9)==0){
      char format[80];
      sscanf(string+9,"%s", format);
      if(strncmp(format, "MUT", 3)==0){
	*EXCHANGE='M';
      }else if(strncmp(format, "FLUX", 4)==0){
	*EXCHANGE='F';
      }else if(strncmp(format, "RATE", 4)==0){
	*EXCHANGE='Q';
      }else if(strncmp(format, "EXCH", 4)==0){
	*EXCHANGE='E';
      }else{
	printf("WARNING, non valid option EXCHANGE= %s\n", format);
	printf("Valid options: FLUX (default), RATE, EXCH, MUT\n");
      }
    }else if(strncmp(string, "MATRIX=", 6)==0){
      char format[80];
      sscanf(string+7,"%s", format);
      if((strncmp(format, "WAG", 3)!=0)&&
	 (strncmp(format, "JTT", 3)!=0)){
	printf("WARNING, matrix %s not defined\n", format);
	printf("Valid options are WAG JTT\n");
	printf("Keeping default %s\n", MATRIX);
      }else{
	strcpy(MATRIX, format);
      }
    }else if(strncmp(string, "TWONUCMUT=", 10)==0){
      sscanf(string+10,"%f", TWONUC);
      if((*TWONUC < 0)||(*TWONUC >1)){
	printf("ERROR, the parameter TWONUCMUT must be set ");
	printf("between zero and one, but it is %.3f\n", *TWONUC);
	exit(8);
      }
      // Output
    }else if(strncmp(string, "kCpG=", 4)==0){
      sscanf(string+5,"%f", kCpG); 
    }else{
      printf("WARNING, uninterpreted line:\n%s", string); 
    }
  }
  fclose(file_in);
  return(1);
}

int Find_string(char *flag, char **argv, int n_arg, int m, char *string){
  int i;
  for(i=0; i<n_arg; i++){
    if(strncmp(argv[i], flag, m)==0){
      i++; if(i<n_arg)strcpy(string, argv[i]);
      return(1);
    }
  }
  return(0);
}

void help(char *prog){
  printf("USAGE %s <input file> ", prog);
  printf("(ex. Prot_evol.in) with default parameters\n");
  printf("ARGUMENTS:\n");
  printf("Protein:\n");
  printf("-pdb <file.pdb>     # Mandatory, unless in input file\n");
  printf("-chain <chain>      # Chain identifier (ex. A)\n");
  printf("-seq  <file_seq>    # file with DNA sequece (optional)\n");
  printf("-ali <file>         # file with protein seq in FASTA format (opt)\n");

  printf("Mean-field computations:\n");
  printf("-meanfield          # Meanfield computation of subst. rates\n");
  printf("-lambda <lambda>    # Parameter for meanfield: l=N_pop*exp(-DG/T)\n");
  printf("                    # If lambda not given, optimal lambda is used\n");
  printf("-model <MODEL>      # Criterion to optimize lambda. Allowed:\n");
  printf("                    # NAT (Max Lik native model)\n");
  printf("                    # ALL (Max Lik misfolding model)\n");
  printf("                    # DG (DeltaG_ALL = target value)\n");

  printf("Mutational model:\n");
  printf("-freq <FREQ>        # Criterion to fit frequencies. Allowed:\n");
  printf("                    # nuc (fit nucleotides from prot sequence)\n");
  printf("                    # aa (fit amino acids from prot sequence)\n");
  printf("                    # input (get nucleotides from input)\n");
  printf("-gc   <GC_bias>     # frequency of nucleotides G+C\n");
  printf("-fG <freq>          # Input frequency for each nucleotide\n");
  printf("-tt   <trans_ratio> # transition-transversion ratio (>1)\n");

  printf("Thermodynamic model:\n");
  printf("-temp <TEMP>        # temperature\n");
  printf("-sU1  <sU1>         # Unfolded entropy per residue\n");

  printf("Evolution simulations:\n");
  printf("-it   <IT_MAX>      # Number of iterations per individual and a.a\n");
  printf("-pop  <N_pop>       # population size\n");


  //printf("-gencode <file>     # file with genetic code\n");
  //printf("-f<n> <freq_nuc>    # frequency of nucleotide n, n=A,T,G,C\n");

  printf("\n\nFORMAT of input file:\n");
  printf("#================================================================\n");
  printf("# A) Input files\n");
  printf("PDB=/data/ortizg/databases/pdb/1tre.pdb  # file_pdb\n");
  printf("CHAIN=  A\n");
  printf("FILE_STR=/home/ubastolla/RESEARCH/PROT_EVOL/INPUT/structures.in\n");
  printf("# List of alternative contact matrices for misfolding computations\n");
  printf("#SEQ=	tpis.dna	# nucleotifde sequence (optional)\n");
  printf("#ALI=	family.aln	# FASTA file with related proteins (optional)\n");
  printf("#================================================================\n");
  printf("# B) Thermodynamic model\n");
  printf("TEMP=	0.5		# Temperature\n");
  printf("SU1=	0.065		# configurational entropy per res (unfold)\n");
  printf("SC1=  0.065		# configurational entropy per res (misfold)\n");
  printf("SC0=  0.0		# configurational entropy offset (misfold)\n");
  printf("REM=   1		# Use up to 1,2,3 moments of misfolding energy\n");
  printf("A_LOC= 1              # Use secondary structure propensities?\n");
  printf("#================================================================\n");
  printf("## C) Mean-field model\n");
  printf("MEANFIELD= 1		# Mean field computations?\n");
  printf("OPT_LAMBDA=1         # Optimize LAMBDA? (1=Yes default)\n");
  printf("LAMBDA= 0.8 		# Value of Lambda if OPT_LAMBDA==0\n");
  printf("MODEL= ALL  		# Type of model. Allowed: ALL (def) NAT\n");
  printf("# D) Simulations of evolution\n");
  printf("TMAX=   000		# ITMAX: # of substitutions\n");
  printf("NEUTRAL= 0		# ");
  printf("#1:Neutral fitness landscape 0: Fitness=1/(1+exp(DG/T))\n");
  printf("NPOP=	10		");
  printf("# effective population size (if MEANFIELD=0, NEUTRAL=0)\n");
  printf("#================================================================\n");
  printf("# E)  Mutation distribution\n");
  printf("GET_FREQ=1		# Allowed: 0,1,2,3\n");
  printf("# 0= Use input mutation parameters\n");
  printf("# 1= Fit mutation parameters from prot sequences\n");
  printf("# 3= Get background distribution from amino acid frequencies\n");
  printf("# 2= Combine fitted mutation model and a.a. frequencies\n");
  printf("# Parameters of the mutation model if GET_FREQ=0:\n");
  printf("REMUT= 1              # Improve mutation par. after MF model?\n");
  printf("FREQ A 0.25		# P(A)\n");
  printf("FREQ T 0.25		# P(T)\n");
  printf("FREQ C 0.25		# P(C)\n");
  printf("FREQ G 0.25		# P(G)\n");
  printf("kCpG=5		# Enhanced rate at CpG dinucleotides\n");
  printf("TT_RATIO=1		# transition-transversion ratio (not CpG)\n");
  printf("TWONUCMUT=0.1		# Ratio between 1-nuc and 2-nuc mutations\n");
  printf("#====================================================================\n");
  printf("# F) Exchangeability matrix\n");
  printf("EXCHANGE=FLUX	     	# Allowed: FLUX (default), RATE, EXCH, MUT\n");
  printf("MATRIX=JTT		# Empirical exchangeability matrix (JTT, WAG)\n");
  printf("#=====================================================================\n");
  printf("# G) Output\n");
  printf("FORMAT=PAML		# Use PAML format for exchangeability matrix\n");
  printf("PRINT_E=0             # Print exchangeability matrix at all sites?\n");
  printf("\n");
  exit(8);
}
